﻿using CAUtilLib;
using CertificateAutorityServer.Assist;
using CertificateAutorityServer.Model.Config;
using CertificateAutorityServer.Model.Rest;
using Microsoft.AspNetCore.Mvc;
using Microsoft.Extensions.Options;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Ocsp;
using Org.BouncyCastle.X509;

namespace CertificateAutorityServer.Service
{
    public class CertificateService
    {
        public CertificateService(
            IOptions<CerServiceConfig> config ,
            ILogger<CertificateService> logger)
        {
            this.config = config.Value;
            Init(config.Value);
        }

        private string RootPath = "/home/cert";

        private const string RcaDirectoryName = "rca";
        private const string TrustedDirectoryName = "tcrt";

        private const string RcaKeyName = "rca_key.pem";
        private const string RcaCertName = "rca.pem";

        private string RcaKeyPath => Path.Combine(RootPath, RcaDirectoryName, RcaKeyName);
        private string RcaCertPath => Path.Combine(RootPath, RcaDirectoryName, RcaCertName);
        private string TrustedCertsPath => Path.Combine(RootPath, TrustedDirectoryName);

        private readonly CerServiceConfig config;

        private void Init(CerServiceConfig value)
        {
            this.RootPath = value.RootPath;
            CheckAndCreateRootCA();
        }

        public void CheckAndCreateRootCA()
        {
            CheckAndCreateFolder();
            if (!File.Exists(RcaKeyPath))
            {
                AsymmetricCipherKeyPair kp = BouncyCastleWrapper.GenerateRsaKeyPair(4096);
                var saveKeyResult = BouncyCastleWrapper.TrySaveAsPemAsync(RcaKeyPath, new object[] { kp.Private }).Result;
                var subject = BouncyCastleWrapper.CreateX509Name(
                    commonName: config.DefaultRcaConfig.CommonName,
                    organizationName: config.DefaultRcaConfig.OrganiaztionName,
                    country: config.DefaultRcaConfig.Country,
                    state: config.DefaultRcaConfig.State);
                X509Certificate cert = BouncyCastleWrapper.GenerateSelfSignedRootCertificate(subject, kp, days: config.DefaultRcaConfig.Days);
                var saveCertResult = BouncyCastleWrapper.TrySaveAsPemAsync(RcaCertPath, new object[] { cert, kp.Private }).Result;
                return;
            }
            if (!File.Exists(RcaCertPath))
            {
                var kp = BouncyCastleWrapper.LoadPemKeyFromFile(RcaKeyPath).Result;
                var subject = BouncyCastleWrapper.CreateX509Name(
                    commonName: config.DefaultRcaConfig.CommonName,
                    organizationName: config.DefaultRcaConfig.OrganiaztionName,
                    country: config.DefaultRcaConfig.Country,
                    state: config.DefaultRcaConfig.State);
                var cert = BouncyCastleWrapper.GenerateSelfSignedRootCertificate(subject, kp, days: config.DefaultRcaConfig.Days);
                var saveCertResult = BouncyCastleWrapper.TrySaveAsPemAsync(RcaCertPath, new object[] { cert, kp.Private }).Result;
            }
        }

        public string GenerateRsaPrivateKeyPemString(int length = 4096)
        {
            var kp = BouncyCastleWrapper.GenerateRsaKeyPair(length);
            return BouncyCastleWrapper.ToStringAsPem(kp).Result;
        }

        public string CreateCertificateReq(CreateCsrReq createCsrReq)
        {
            var kp = BouncyCastleWrapper.LoadPemKeyFromString(createCsrReq.Key);
            var csr = BouncyCastleWrapper.CreateCertificateReq(kp, createCsrReq.CommonName, createCsrReq.OrganizationName);
            return BouncyCastleWrapper.ToStringAsPem(csr).Result;
        }

        public async Task<string> GetRcaAsPem()
        {
            var cert = await BouncyCastleWrapper.LoadPemCertFromFile(RcaCertPath);
            return await BouncyCastleWrapper.ToStringAsPem(cert);
        }

        public async Task<string> SignCsrByRca(SignCsrReq signCsrReq)
        {
            var cert = await BouncyCastleWrapper.LoadPemCertFromFile(RcaKeyPath);
            var rcaKey = await BouncyCastleWrapper.LoadPemKeyFromFile(RcaKeyPath);
            var csr = BouncyCastleWrapper.LoadPemCsrFromString(signCsrReq.Csr);
            //var singedCert = BouncyCastleWrapper.SignCertificate(csr, cert, rcaKey);
            var singedCert = BouncyCastleWrapper.GenerateEndCertificate(csr, cert, rcaKey);
            return await BouncyCastleWrapper.ToStringAsPem(singedCert);
        }

        public async Task<bool> CheckCsrByRca(string crt)
        {
            var cert = BouncyCastleWrapper.LoadPemCertFromString(crt);
            var rcaKey = await BouncyCastleWrapper.LoadPemKeyFromFile(RcaKeyPath);
            return BouncyCastleWrapper.ValidateCert(cert, rcaKey.Public);
        }

        public void ReGenerateRca(bool addTrusted = true)
        {
            var cert = BouncyCastleWrapper.LoadPemCertFromFile(RcaCertPath).Result;
            if (addTrusted)
            {
                AddTrustedCrt(cert);
            }

            if (File.Exists(RcaKeyPath))
            {
                File.Delete(RcaKeyPath);
            }
            if (File.Exists(RcaCertPath))
            {
                File.Delete(RcaCertPath);
            }
            CheckAndCreateRootCA();
        }

        public void AddTrustedCrt(string crtString)
        {
            var crt = BouncyCastleWrapper.LoadPemCertFromString(crtString);
            AddTrustedCrt(crtString);
        }

        public void AddTrustedCrt(X509Certificate crt)
        {
            var path = Path.Combine(TrustedCertsPath,  crt.SerialNumber.ToString());
            BouncyCastleWrapper.TrySaveAsPemAsync(path, new object[] { crt }).Wait();
        }

        public List<string> GetTrustedCrts()
        {
            var files = Directory.GetFiles(TrustedCertsPath).Select(x=> Path.GetFileName(x)).ToList();
            return files;
        }

        public Task<bool> VerifyCrtSignedByTrused(string crtString)
        {
            var crt = BouncyCastleWrapper.LoadPemCertFromString(crtString);
            return VerifyCrtSignedByTrused(crtString);
        }

        public async Task<bool> VerifyCrtSignedByTrused(X509Certificate crt)
        {
            var trustedCerts = Directory.GetFiles(TrustedCertsPath);
            foreach (var trustedCert in trustedCerts)
            {
                var cert = await BouncyCastleWrapper.LoadPemCertFromFile(trustedCert);
                var checkResult = BouncyCastleWrapper.ValidateCert(crt, cert.GetPublicKey());
                if (checkResult)
                {
                    return true;
                }
            }
            return false;
        }

        private void CheckAndCreateFolder()
        {
            if (!Directory.Exists(config.RootPath))
            {
                Directory.CreateDirectory(config.RootPath);
            }
            var caFolder = Path.Combine(config.RootPath, CaUtil.RcaDirectoryName);
            if (!Directory.Exists(caFolder))
            {
                Directory.CreateDirectory(caFolder);
            }
            var tcFolder = Path.Combine(config.RootPath, CaUtil.TrustedDirectoryName);
            if (!Directory.Exists(tcFolder))
            {
                Directory.CreateDirectory(tcFolder);
            }
        }
    }
}
